#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "cptutils/comments.h"
#include "cptutils/btrace.h"

#include <stdbool.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <time.h>
#include <ctype.h>

typedef struct line_t line_t;

struct line_t
{
  char *text;
  line_t *next;
};

struct comments_t
{
  bool finalised;
  size_t offset;
  line_t *line;
};

comments_t* comments_new(void)
{
  comments_t *comments;

  if ((comments = malloc(sizeof(comments_t))) == NULL)
    return NULL;

  comments->finalised = false;
  comments->line = NULL;

  return comments;
}

static void line_destroy(line_t *line)
{
  if (line != NULL)
    {
      line_destroy(line->next);
      free(line->text);
      free(line);
    }
}

void comments_destroy(comments_t *comments)
{
  if (comments != NULL)
    {
      line_destroy(comments->line);
      free(comments);
    }
}

static int line_each(const line_t *line,
                     size_t offset,
                     int (*f)(const char*, void*),
                     void *arg)
{
  if (line == NULL)
    return 0;

  return
    line_each(line->next, offset, f, arg) ||
    f(line->text ? line->text + offset : NULL, arg);
}

int comments_each(const comments_t *comments,
                  int (*f)(const char*, void*),
                  void *arg)
{
  return line_each(comments->line, comments->offset, f, arg);
}

static int line_count(const char *text, void *arg)
{
  size_t *n = arg;
  (*n)++;
  return 0;
}

size_t comments_count(const comments_t *comments)
{
  size_t n = 0;
  comments_each(comments, line_count, &n);
  return n;
}

static int line_copy(const char *text, void *arg)
{
  comments_t *dst = arg;
  return comments_push(dst, text ? text : "");
}

int comments_copy(const comments_t *src, comments_t *dst)
{
  dst->finalised = src->finalised;
  dst->offset = 0;
  return comments_each(src, line_copy, dst);
}

int comments_cat(const comments_t **src,
                 size_t n,
                 const char *sep,
                 comments_t *dst)
{
  if (n == 0)
    return 0;

  if (comments_each(src[0], line_copy, dst) != 0)
    {
      btrace("concatenate first");
      return 1;
    }

  for (size_t i = 1 ; i < n ; i++)
    {
      if ((line_copy(sep, dst) != 0) ||
          (comments_each(src[i], line_copy, dst) != 0))
        {
          btrace("conatenate item %zi", i);
          return 1;
        }
    }

  dst->finalised = true;
  dst->offset = 0;

  return 0;
}

static void chomp(char *text)
{
  char *c;

  if ((c = strchr(text, '\n')) != NULL)
    *c = '\0';
}

static bool string_is_whitespace(const char *string)
{
  for (const char *c = string ; *c != '\0' ; c++)
    if (! isspace(*c))
      return false;

  return true;
}

int comments_push(comments_t *comments, const char *text)
{
  int err = 1;
  line_t *line;

  if ((line = malloc(sizeof(line_t))) != NULL)
    {
      if (string_is_whitespace(text))
        {
          line->text = NULL;
          line->next = comments->line;
          comments->line = line;
          err = 0;
        }
      else
        {
          char *text_copy = NULL;

          if ((text_copy = strdup(text)) == NULL)
            free(line);
          else
            {
              chomp(text_copy);
              line->text = text_copy;
              line->next = comments->line;
              comments->line = line;
              err = 0;
            }
        }
    }

  return err;
}

static void comments_line_clear(comments_t *comments)
{
  line_destroy(comments->line);
  comments->line = NULL;
  comments->offset = 0;
}

static int line_write(const char *text, void *arg)
{
  FILE *st = arg;
  int err;

  if (text == NULL)
    err = fprintf(st, "\n");
  else
    err = fprintf(st, "%s\n", text);

  return (err < 0) ? 1 : 0;
}

static int comments_write(const comments_t *comments, FILE *st)
{
  if (comments->finalised == false)
    {
      btrace("comments not finalised");
      return 1;
    }
  else
    return comments_each(comments, line_write, st);
}

static int line_offset(const char *text, void *arg)
{
  if (text == NULL)
    return 0;

  size_t n = 0, *nmax = arg;

  for (const char *c = text ; (*c != '\0') && isspace(*c) ; c++)
    n++;

  if (n < *nmax)
    *nmax = n;

  return 0;
}

static int comments_offset(comments_t *comments)
{
  size_t offset = SIZE_MAX;

  if (line_each(comments->line, 0, line_offset, &offset) != 0)
    {
      btrace("failed to get comment offset");
      return 1;
    }

  if (offset == SIZE_MAX)
    offset = 0;

  comments->offset = offset;

  return 0;
}

static int lines_header_trim(line_t *line)
{
  if (line == NULL)
    return 0;

  if (lines_header_trim(line->next) == 0)
    {
      line_t *line_next = line->next;

      if ((line_next != NULL) &&
          (line_next->next == NULL) &&
          (line_next->text == NULL))
        {
          free(line_next);
          line->next = NULL;
        }
    }

  return 0;
}

static int comments_header_trim(comments_t *comments)
{
  return lines_header_trim(comments->line);
}

static int comments_footer_trim(comments_t *comments)
{
  line_t *line;

  while (((line = comments->line) != NULL) &&
         (line->text == NULL))
    {
      comments->line = line->next;
      free(line);
    }

  return 0;
}

int comments_finalise(comments_t *comments)
{
  if ((comments_offset(comments) == 0) &&
      (comments_header_trim(comments) == 0) &&
      (comments_footer_trim(comments) == 0))
    {
      comments->finalised = true;
      return 0;
    }
  else
    return 1;
}

static int comments_generate(comments_t *comments)
{
  size_t n = 256;
  char line[n];
  time_t t = time(NULL);

  if (snprintf(line, n, "cptutils %s, %s", VERSION, ctime(&t)) > n)
    {
      btrace("generated comment too large for buffer");
      return 1;
    }

  if (comments_push(comments, line) != 0)
    {
      btrace("failed comment push");
      return 1;
    }

  return 0;
}

static int comments_read(comments_t *comments, FILE *stream)
{
  size_t n = 256;
  char line[n];

  while (fgets(line, n, stream) != NULL)
    {
      if (comments_push(comments, line) != 0)
        {
          btrace("failed comment push");
          return 1;
        }
    }

  return 0;
}

/*
  this is the main comment-handling function, it will typically
  be called between read and write, and is passed the comments as
  determined by read (in which case it may write those commants to
  file).  Then (depending on options) we may replace the comments
  by generation or read from file, which the program's writer can
  the add to the output.  This modifies the first argument.
*/

int comments_process(comments_t *comments, const comments_opt_t *opt)
{
  if (opt->path.output != NULL)
    {
      FILE *st;
      int err = 1;

      if ((st = fopen(opt->path.output, "w")) != NULL)
        {
          err = comments_write(comments, st);
          fclose(st);
        }

      if (err != 0)
        {
          btrace("failed comment write");
          return 1;
        }
    }

  if (opt->action != action_retain)
    {
      comments_line_clear(comments);

      if (opt->action == action_generate)
        {
          if (comments_generate(comments) != 0)
            {
              btrace("failed comment generate");
              return 1;
            }
        }
      else if (opt->action == action_read)
        {
          FILE *st;
          int err = 1;

          if ((st = fopen(opt->path.input, "r")) != NULL)
            {
              err = comments_read(comments, st);
              fclose(st);
            }

          if (err != 0)
            {
              btrace("failed comment read from '%s'", opt->path.input);
              return 1;
            }
        }
    }

  return 0;
}
