#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "qgssvg.h"

#include <cptutils/qgs-list.h>
#include <cptutils/qgs-list-read.h>
#include <cptutils/svg-list.h>
#include <cptutils/svg-list-write.h>
#include <cptutils/btrace.h>
#include <cptutils/macro.h>

#include <string.h>

static double clamp(double min, double max, double v)
{
  if (v > max)
    return max;
  if (v < min)
    return min;
  return v;
}

static double qgssvg_value(double v)
{
  return clamp(0, 100, 100 * v);
}

static double qgssvg_opacity(double v)
{
  return clamp(0, 1, v / 255);
}

static int qgssvg_one(const qgs_t *qgs, svg_t *svg)
{
  switch (qgs->type)
    {
    case QGS_TYPE_UNSET:
    case QGS_TYPE_DISCRETE:
      btrace("QGS type %i not handled", qgs->type);
      return 1;
    case QGS_TYPE_INTERPOLATED:
      break;
    default:
      btrace("QGS type %i unknown", qgs->type);
    }

  if (qgs->n == 0)
    {
      btrace("QGS has no stops");
      return 1;
    }

  for (size_t i = 0 ; i < qgs->n ; i++)
    {
      const qgs_entry_t *entry = qgs->entries + i;
      svg_stop_t stop = {
        .colour = entry->rgb,
        .value = qgssvg_value(entry->value),
        .opacity = qgssvg_opacity(entry->opacity)
      };

      if (svg_append(stop, svg) != 0)
        {
          btrace("failed to append stop %zu", i);
          return 1;
        }
    }

  strncpy((char*)(svg->name), qgs->name, SVG_NAME_LEN - 1);

  if (comments_copy(qgs->comments, svg->comments) != 0)
    {
      btrace("failed to copy comments");
      return 1;
    }

  return 0;
}

static int qgssvg_iterand(qgs_t *qgs, void *arg)
{
  svg_list_t *svg_list = arg;
  svg_t *svg;
  int err = 1;

  if ((svg = svg_list_next(svg_list)) != NULL)
    {
      if (qgssvg_one(qgs, svg) == 0)
        err = 0;
      else
        {
          if (svg_list_revert(svg_list) == 0)
            err = 0;
          else
            btrace("failed svg-list revert");
        }
    }
  else
    btrace("failed svg-list next");

  return err;
}

static svg_list_t* qgssvg_convert(qgs_list_t *qgs)
{
  svg_list_t *svg;

  if ((svg = svg_list_new()) != NULL)
    {
      if (qgs_list_iterate(qgs, qgssvg_iterand, svg) == 0)
        return svg;

      svg_list_destroy(svg);
    }

  return NULL;
}

static int qgq_list_comments_iterand(qgs_t *qgs, void *arg)
{
  const qgssvg_opt_t *opt = arg;
  return comments_process(qgs->comments, &(opt->comments));
}

static int qgs_list_comments(qgs_list_t *qgs, const qgssvg_opt_t *opt)
{
  return qgs_list_iterate(qgs, qgq_list_comments_iterand, (void*)opt);
}

int qgssvg(const qgssvg_opt_t *opt)
{
  qgs_list_t *qgs_list;
  int err = 1;

  if ((qgs_list = qgs_list_read(opt->input.path)) != NULL)
    {
      if (qgs_list_comments(qgs_list, opt) == 0)
        {
          svg_list_t *svg_list;

          if ((svg_list = qgssvg_convert(qgs_list)) != NULL)
            {
              size_t n = svg_list_size(svg_list);

              if (n > 0)
                {
                  const char *path = opt->output.path;
                  svg_preview_t preview = { .use = false };

                  if (svg_list_write(svg_list, path, &preview) == 0)
                    {
                      err = 0;
                      if (opt->verbose)
                        printf("wrote %zu gradients to %s\n", n, path);
                    }
                  else
                    btrace("failed write to %s", IFNULL(path, "<stdout>"));

                  svg_list_destroy(svg_list);
                }
              else
                btrace("no gradients converted");
            }
          else
            btrace("failed convert");
        }
      else
        btrace("failed to process comments");

      qgs_list_destroy(qgs_list);
    }
  else
    btrace("failed read of %s", opt->input.path);

  return err;
}
